--
-- DMX512 receiver
--

library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity data_receiver is
    Port ( clk      : in std_logic;                     -- system clock, 8MHz
           rx_data  : in std_logic;                     -- DMX512 data
           chn_sel  : in std_logic;                     -- select DMX512 channels
           pwm_data : out std_logic_vector(7 downto 0); -- parallel DMX data
           pwm_adr  : out std_logic_vector(7 downto 0); -- address for PWM RAM
           page     : out std_logic;                    -- RAM page selector
           trigger  : out std_logic;                    -- trigger new MUX cycle
           we       : out std_logic);                   -- write enable
end data_receiver;

architecture Behavioral of data_receiver is

type state_t is (idle, rx_run, rx_eval, rx_end);

signal rx_smp       : std_logic_vector(3 downto 0);     -- sample and delay rx_data
signal shift        : std_logic_vector(9 downto 0);     -- rx shift receiver, 8+2
signal smp_cnt      : std_logic_vector(3 downto 0);     -- bit timing counter
signal bit_cnt      : std_logic_vector(3 downto 0);     -- bit counter
signal state        : state_t;                          -- state for FSM
signal adr_cnt      : std_logic_vector(7 downto 0);     -- Address counter for RAM
signal dmx_cnt      : std_logic_vector(9 downto 0);     -- DMX channel counter
signal rx_filt      : std_logic;                        -- filtered rx_data
signal page_int     : std_logic;                        -- RAM page signal
signal ce           : std_logic;                        -- clock enable for UART
signal adr_cnt_ce   : std_logic;                        -- clock enable for RAM address counter

begin

-- signal assignment

pwm_adr  <= adr_cnt;
page     <= page_int;

-- clock enable for UART
-- master clock is 8 MHz, DMX baud rate is 250 kb/s
-- -> Oversampling wih 4MHz

process(clk)
begin
  if rising_edge(clk) then
    ce <= not ce;               -- simple 2:1 divider
  end if;
end process;

-- sample rx_data

process(clk)
begin
  if rising_edge(clk) then
    if ce='1' then
      rx_smp <= rx_smp(2 downto 0) & rx_data;           -- two sample, two delay stages
    end if;
  end if;
end process;

-- filter rx_data, majority vote out of three

process(rx_smp)
begin
  case rx_smp(3 downto 1) is
     when "000" => rx_filt <= '0';
     when "001" => rx_filt <= '0';
     when "010" => rx_filt <= '0';
     when "011" => rx_filt <= '1';
     when "100" => rx_filt <= '0';
     when "101" => rx_filt <= '1';
     when "110" => rx_filt <= '1';
     when "111" => rx_filt <= '1';
     when others => null;
  end case;
end process;

-- bit timing counter

process(clk)
begin
  if rising_edge(clk) then
    if ce='1' then
      if rx_smp(2 downto 1) = "10" and state=idle then  -- waiting for falling edge 
        smp_cnt <= "0000";      -- synchronize
      else
        smp_cnt <= smp_cnt+1;
      end if;
    end if;
  end if;
end process;

-- clock enable for address counter

process(clk)
begin
  if rising_edge(clk) then
    if chn_sel = '0' then         -- board uses DMX-channel 1..243
      if dmx_cnt =0 then
        adr_cnt_ce  <= '1';
      elsif dmx_cnt =244 then
        adr_cnt_ce  <= '0';
      end if;
    else                          -- board uses DMX-channel 244..486
      if dmx_cnt =243 then
        adr_cnt_ce  <= '1';
      elsif dmx_cnt =487 then
        adr_cnt_ce  <= '0';
      end if; 
    end if;
  end if;
end process;

-- FSM
-- DMX channels 1..243 are stored in RAM adress 0..242!
-- DMX channel 0 is stored in in RAM address 255! not used now
-- all DMX channels > 243 are dumped into RAM address 243

process(clk)
begin
  if rising_edge(clk) then
    case state is 
      when idle =>
        bit_cnt <= "0000";
        trigger <= '0';
        if ce='1' and rx_smp(2 downto 1) = "10" then    -- falling edge, start bit 
          state <= rx_run;
        end if;

      when rx_run => 
        if ce='1' and smp_cnt=10 then
          bit_cnt <= bit_cnt +1;
          shift <= rx_filt & shift(9 downto 1);     -- shift in filtered rx_data, LSB first 
          if bit_cnt=10 then
            state <= rx_eval;
          end if;
        end if;

      when rx_eval =>                   -- rx complete, evaluate data
        pwm_data <= shift(7 downto 0);  --copy data, ease simulation 
        if shift=0 then                 -- all zero -> BREAK
          dmx_cnt  <= (others =>'0');   -- reset channel counter
          adr_cnt  <= "11111111";       -- new frame, reset address counter
          page_int <= not page_int;     -- swap RAM page
          trigger  <= '1';              -- trigger new MUX cycle in synchronous mode
          state    <= idle;
        else
          if shift(9 downto 8)="11" then
            we <= '1';                  -- write data only, if both stop bits are correct
          end if;
          state <= rx_end;
        end if;

      when rx_end =>
        dmx_cnt <= dmx_cnt +1;          -- count
        if adr_cnt_ce = '1' then        -- 
          adr_cnt <= adr_cnt +1;        -- increment address counter
        end if;
        we <= '0';
        trigger <= '0';        
        state <= idle;

    end case;
  end if;
end process;

end Behavioral;
